"""Encoders for caching data sources."""

from tacobi.data_source.encode.base import EncodedDataType, Encoder
from tacobi.data_source.encode.polars import PolarsEncoder
from tacobi.data_source.encode.pydantic import PydanticEncoder

__all__ = ["EncodedDataType", "Encoder", "PolarsEncoder", "PydanticEncoder"]



"""Encoders for Pydantic models."""

from dataclasses import dataclass
from typing import TypeVar

from pydantic import BaseModel

from tacobi.data_model.models import DataModelType
from tacobi.data_source.encode import EncodedDataType, Encoder

BaseModelType = TypeVar("BaseModelType", bound=BaseModel)


@dataclass
class PydanticEncoder(Encoder):
    """An encoder that can encode and decode Pydantic models."""

    base_model: BaseModelType
    """ The base model that is used to encode and decode the data. """

    def encode(self, data: DataModelType) -> EncodedDataType:
        """Encode the data."""
        return data.model_dump_json().encode("utf-8")

    def decode(self, data: EncodedDataType) -> DataModelType:
        """Decode the data."""
        return self.base_model.model_validate_json(data.decode("utf-8"))



"""Encoders for Polars data."""

import io
from dataclasses import dataclass

import polars as pl

from tacobi.data_source.encode import EncodedDataType, Encoder


@dataclass
class PolarsEncoder(Encoder):
    """An encoder that can encode and decode Polars data."""

    def encode(self, data: pl.LazyFrame) -> EncodedDataType:
        """Encode the data."""
        buffer = io.BytesIO()
        data.sink_parquet(buffer)
        return buffer.getvalue()

    def decode(self, data: EncodedDataType) -> pl.LazyFrame:
        """Decode the data."""
        return pl.scan_parquet(io.BytesIO(data))



"""The main app class for TacoBI."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import TypeVar
from uuid import UUID

from apscheduler.triggers.base import BaseTrigger

from tacobi.data_model.models import DataModelType
from tacobi.data_source import CachedDataSource, DataSourceManager
from tacobi.view import MaterializedView, View, ViewManager

T = TypeVar("T", bound=Callable[[DataModelType], Awaitable[DataModelType]])


@dataclass
class TacoBIApp:
    """The main app class for TacoBI."""

    view_manager: ViewManager
    """ Manager used for scheduling views."""

    data_source_manager: DataSourceManager = field(default_factory=DataSourceManager)
    """ Manager used for scheduling data sources."""

    _view_function_ids: dict[Callable, UUID] = field(default_factory=dict)
    """ A dictionary of view functions. """

    # Data Source Management

    def data_source(
        self,
        name: str,
        trigger: BaseTrigger,
    ) -> Callable[
        [Callable[[DataModelType | None], Awaitable[DataModelType]]],
        Callable[[], DataModelType | None],
    ]:
        """Register a data source.

        Also transforms the function into a synchronous getter of the latest data.

        ### Arguments:
        - name: The name of the data source.
        - trigger: The trigger that will be used to schedule the data source.

        ### Returns:
        A function that returns the latest data from the data source.
        """

        def wrapper(
            func: Callable[[DataModelType | None], Awaitable[DataModelType]],
        ) -> Callable[[], DataModelType | None]:
            data_source = CachedDataSource(name=name, function=func, trigger=trigger)
            self.data_source_manager.add_data_source(data_source)

            return data_source.get_latest_data

        return wrapper

    # View Management

    def view(
        self,
        name: str,
        route: str | None = None,
        dependencies: list[Callable] | None = None,
    ) -> Callable[[T], T]:
        """Register a view.

        ### Arguments:
        - name: The name of the view.
        - route: The route of the view.
        - dependencies: The dependencies of the view.

        ### Returns:
        The view function itself.
        """

        def wrapper(
            func: Callable[[T], Awaitable[DataModelType]],
        ) -> Callable[[T], Awaitable[DataModelType]]:
            # Get the dependencies
            try:
                dep_ids = (
                    [self._view_function_ids[dep] for dep in dependencies]
                    if dependencies
                    else []
                )
            except KeyError as e:
                msg = f"Dependency {e} not found"
                raise ValueError(msg) from e

            # Add the view to the view manager
            view = View(
                name=name,
                function=func,
                route=route,
                dependencies=dep_ids,
            )
            self.view_manager.add_view(view)

            # Register the view function in the view manager
            self._view_function_ids[func] = view.id
            return func

        return wrapper

    def materialized_view(
        self,
        name: str,
        route: str | None = None,
        dependencies: list[Callable] | None = None,
    ) -> Callable[
        [Callable[[DataModelType | None], Awaitable[DataModelType]]],
        Callable[[], DataModelType | None],
    ]:
        """Register a materialized view.

        Also transforms the function into a synchronous getter of the latest data.

        ### Arguments:
        - route: The route of the view.
        - dependencies: The dependencies of the view.

        ### Returns:
        The view function itself.
        """

        def wrapper(
            func: Callable[[DataModelType | None], Awaitable[DataModelType]],
        ) -> Callable[[], DataModelType | None]:
            # Get the dependencies
            try:
                dep_ids = (
                    [self._view_function_ids[dep] for dep in dependencies]
                    if dependencies
                    else []
                )
            except KeyError as e:
                msg = f"Dependency {e} not found"
                raise ValueError(msg) from e

            # Add the view to the view manager
            view = MaterializedView(
                name=name,
                function=func,
                route=route,
                dependencies=dep_ids,
            )
            self.view_manager.add_materialized_view(view)

            # Register the view function in the view manager
            self._view_function_ids[func] = view.id

            # Return the latest data from the view
            def get_latest_view_data() -> DataModelType | None:
                return view.get_latest_data()

            return get_latest_view_data

        return wrapper

    # Lifecycle

    async def start(self) -> None:
        """Start the recomputation of datasets and materialized views."""
        await self.data_source_manager.start()
        self.view_manager.start()

    async def stop(self) -> None:
        """Stop the recomputation of datasets and materialized views."""
        await self.data_source_manager.stop()
        self.view_manager.stop()



"""Cache backend for data sources."""

from tacobi.data_source.cache.base import CacheBackend, EncodedDataType
from tacobi.data_source.cache.sqlite import SQLiteCache

__all__ = ["CacheBackend", "SQLiteCache", "EncodedDataType"]



"""Data sources are functions that are called to update the data.

They can be scheduled to be called at a specific time and can be cached to avoid
redundant calls.
"""

from tacobi.data_source.cache import CacheBackend, SQLiteCache
from tacobi.data_source.models import (
    CachedDataSource,
    DataSourceManager,
)

__all__ = [
    "CachedDataSource",
    "CacheBackend",
    "DataSourceManager",
    "SQLiteCache",
]



"""Type utilities for extracting types from type annotations."""

import inspect
from collections.abc import Callable
from typing import TypeVar, get_args, get_origin, get_type_hints

T = TypeVar("T")
U = TypeVar("U")


def extract_return_type(func: Callable) -> type:
    """Extract the return type annotation from a function.

    ### Arguments:
    - func: The function to inspect

    ### Returns:
    The return type annotation

    ### Raises:
    - TypeError: If the function has no return type annotation
    """
    return_type = get_type_hints(func).get("return")
    if return_type is None:
        msg = "Function has no return type annotation"
        raise TypeError(msg)
    return return_type


def extract_container_type(container_type: type, expected_container_type: type) -> type:
    """Extract and validate the container type from a type annotation.

    ### Arguments:
    - container_type: The container type annotation
    - expected_container_type: The expected container type (e.g., DataFrame)

    ### Returns:
    The validated container type

    ### Raises:
    - TypeError: If the container type is not the expected type
    """
    origin = get_origin(container_type)
    if not isinstance(origin, type(expected_container_type)):
        msg = f"Expected {expected_container_type.__name__}, got {origin}"
        raise TypeError(msg)
    return origin


def extract_container_content(
    container_type: type, expected_container_type: type, content_type: type
) -> type:
    """Extract the content type from a container type annotation.

    ### Arguments:
    - container_type: The container type annotation
    - expected_container_type: The expected container type (e.g., DataFrame)
    - content_type: The expected base class of the content (e.g., DataFrameModel)

    ### Returns:
    The content type used in the container type annotation

    ### Raises:
    - TypeError: If the container type is not properly formatted
    """
    # Check if it's the expected container type
    extract_container_type(container_type, expected_container_type)

    # Extract the model from Container[Model]
    type_args = get_args(container_type)
    if not type_args or len(type_args) != 1:
        msg = f"{expected_container_type.__name__} must have exactly one type argument"
        raise TypeError(msg)

    content_class = type_args[0]

    # Verify it's the expected content type
    if not (inspect.isclass(content_class) and issubclass(content_class, content_type)):
        msg = f"Expected {content_type.__name__}, got {content_class}."
        raise TypeError(msg)

    return content_class


def extract_model_from_typehints(
    func: Callable, expected_container_type: type, expected_model_base: type
) -> type:
    """Extract a model class from a function that returns Container[Model].

    ### Arguments:
    - func: The function to inspect
    - expected_container_type: The expected container type (e.g., DataFrame)
    - expected_model_base: The expected base class of the model (e.g., DataFrameModel)

    ### Returns:
    The model class used in the container type annotation

    ### Raises:
    - TypeError: If the function's return type is not properly annotated
    """
    return_type = extract_return_type(func)
    return extract_container_content(
        return_type, expected_container_type, expected_model_base
    )



"""The main app class for TacoBI."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import TypeVar

import rustworkx as rx
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.base import BaseTrigger
from fastapi import FastAPI

from tacobi.data_model.models import DataModelType
from tacobi.view import BaseView, MaterializedView, View
from tacobi.view.dataset_schema import Schema

T = TypeVar("T", bound=Callable[[DataModelType], Awaitable[DataModelType]])


@dataclass
class ViewManager:
    """The main app class for TacoBI."""

    recompute_trigger: BaseTrigger
    """ The trigger that will be used to recompute the materialized views. """

    fastapi_app: FastAPI
    """ The FastAPI app that will be used to serve the views. """

    _recompute_scheduler: AsyncIOScheduler = field(default_factory=AsyncIOScheduler)
    """ The scheduler that will be used to recompute the materialized views. """

    _views: list[View] = field(default_factory=list)
    """ The views that are used in the app. """

    _materialized_views: list[MaterializedView] = field(default_factory=list)
    """ The materialized views that are used in the app. """

    # View Management

    def add_view(self, view: View) -> None:
        """Add a view to the app."""
        self._views.append(view)
        if view.route:
            self._attach_view_to_fastapi(view)

    def add_materialized_view(self, view: MaterializedView) -> None:
        """Add a materialized view to the app."""
        self._materialized_views.append(view)
        if view.route:
            self._attach_materialized_view_to_fastapi(view)

    def _get_sorted_views(self) -> list[BaseView]:
        """Get all views sorted by dependency order.

        ### Returns:
        List of views in order of calculation.
        """
        # Combine all views
        all_views = self._views + self._materialized_views

        # Create a directed graph
        graph = rx.PyDiGraph()

        # Add nodes and store mapping of UUID to node index
        node_map = {view.id: graph.add_node(view) for view in all_views}

        # Add edges for dependencies
        for view in all_views:
            if view.dependencies:
                for dep_view in view.dependencies:
                    graph.add_edge(node_map[dep_view], node_map[view.id], None)

        try:
            # Get topological sort
            sorted_indices = rx.topological_sort(graph)
            return [graph[node_idx] for node_idx in sorted_indices]
        except rx.DAGHasCycle as e:
            msg = "Circular dependency detected in views"
            raise ValueError(msg) from e

    async def _recompute_materialized_views(self) -> None:
        """Recompute all materialized views in dependency order."""
        materialized_views = [
            v for v in self._get_sorted_views() if isinstance(v, MaterializedView)
        ]

        print(f"Recomputing {len(materialized_views)} materialized views:")

        for i, view in enumerate(materialized_views):
            print(f"{i + 1}. Recomputing {view.name}...")
            await view.recompute_latest_data()

    # Lifecycle

    def start(self) -> None:
        """Start the recomputation of materialized views."""
        self._recompute_scheduler.add_job(
            self._recompute_materialized_views,
            trigger=self.recompute_trigger,
        )
        print("Starting scheduler")
        self._recompute_scheduler.start()

    def stop(self) -> None:
        """Stop the recomputation of materialized views."""
        self._recompute_scheduler.shutdown()

    # REST API

    def _attach_materialized_view_to_fastapi(
        self,
        view: MaterializedView,
    ) -> None:
        """Attach a materialized view to a FastAPI route.

        ### Arguments:
        - view: The materialized view to attach to the FastAPI route.
        """
        self.fastapi_app.get(view.route, response_model=view.fastapi_response_model)(
            view.get_latest_data
        )

    def _attach_view_to_fastapi(self, view: View) -> None:
        """Attach a view to a FastAPI route.

        ### Arguments:
        - view: The view to attach to the FastAPI route.
        """
        self.fastapi_app.get(view.route, response_model=view.fastapi_response_model)(
            view.function
        )

    # Schema

    @property
    def schema(self) -> Schema:
        """The schema of the exposed view endpoints."""
        all_views = self._views + self._materialized_views
        exposed_views = [v for v in all_views if v.route]
        return Schema(
            views=[v.schema for v in exposed_views],
        ).json_schema



"""Models for data models."""

from datetime import datetime
from typing import TypeVar

from pandera.typing.common import DataFrameBase
from pydantic import BaseModel


DataModel = DataFrameBase | BaseModel
"""A valid data model can either be a Pandera Polars `DataFrameModel` or a
Pydantic `BaseModel`. Views and data sources can return either of these types.
"""

DataModelType = TypeVar("DataModelType", bound=DataModel)
"""A type variable for the data model type."""



"""Base view class."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, Generic
from uuid import UUID, uuid4

from pandera import DataFrameModel
from pandera.typing import DataFrame
from pydantic import BaseModel

from tacobi.data_model.models import DataModelType
from tacobi.view.dataset_schema import (
    ViewEndpointSchema,
    create_column_base_model_from_dataframe_model,
)
from tacobi.view.type_utils import (
    extract_container_type,
    extract_model_from_typehints,
    extract_return_type,
)


@dataclass
class BaseView(Generic[DataModelType]):
    """Base view class."""

    function: Callable[..., Awaitable[DataModelType]]
    """The function to call to update the view."""

    dependencies: list[UUID] = field(default_factory=list)
    """The dependencies of the view used for topological sorting."""

    route: str | None = None
    """The optional REST route of the view."""

    name: str | None = None
    """The optional name of the view for logging and debugging."""

    id: UUID = field(default_factory=uuid4)
    """The unique identifier for the view."""

    def __str__(self) -> str:
        """Get the string representation of the view."""
        return f"View(name={self.name}, id={self.id})"

    def __hash__(self) -> int:
        """Hash the view."""
        return hash(self.id)

    @cached_property
    def return_type(self) -> type[DataModelType]:
        """The return type of the view."""
        return extract_return_type(self.function)

    @cached_property
    def input_type(self) -> type[BaseModel] | None:
        """The input parameter of the view."""
        input_params = list(self.function.__annotations__.items())[1:]

        if len(input_params) == 0:
            return None

        if len(input_params) > 1:
            msg = f"""View must have exactly one input parameter. Input params:
            {input_params}"""
            raise ValueError(msg)

        _, param_type = input_params[0]

        if not issubclass(param_type, BaseModel):
            msg = f"""Input parameter must be a Pydantic BaseModel. Input param:
            {param_type}"""
            raise TypeError(msg)

        return param_type

    @cached_property
    def fastapi_response_model(self) -> type[BaseModel]:
        """The response model for the view."""
        return_type = self.return_type

        # If the return type is a BaseModel, we just return it as is
        if issubclass(return_type, BaseModel):
            return return_type

        # Return a FastAPI model configured for converting to a dict
        # Read more here: https://pandera.readthedocs.io/en/latest/fastapi.html

        container_type = extract_container_type(return_type, DataFrame)
        model = extract_model_from_typehints(
            self.function, container_type, DataFrameModel
        )

        class FastAPIModel(model):
            class Config:
                to_format: ClassVar[str] = "dict"
                to_format_kwargs: ClassVar[dict[str, str]] = {"orient": "records"}

        return container_type[FastAPIModel]

    @cached_property
    def output_schema(self) -> type[BaseModel]:
        """The output schema of the view."""
        return_type = self.return_type

        if issubclass(return_type, BaseModel):
            return return_type

        container_type = extract_container_type(return_type, DataFrame)
        model = extract_model_from_typehints(
            self.function, container_type, DataFrameModel
        )

        return create_column_base_model_from_dataframe_model(model)

    @cached_property
    def schema(self) -> ViewEndpointSchema:
        """The schema of the view."""
        input_schema = self.input_type
        data_schema = self.output_schema

        return ViewEndpointSchema(
            route=self.route,
            input_schema=input_schema,
            data_schema=data_schema,
        )



"""Models for data sources."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass, field
from typing import Generic

import polars as pl
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.base import BaseTrigger
from pydantic import BaseModel

from tacobi.data_model.models import DataModelType
from tacobi.data_source.cache import CacheBackend, SQLiteCache
from tacobi.data_source.encode import Encoder, PolarsEncoder, PydanticEncoder


@dataclass
class CachedDataSource(Generic[DataModelType]):
    """A data source that is locally stored and can be instantly queried."""

    name: str
    """ The *UNIQUE* name of the data source. """

    function: Callable[[DataModelType | None], Awaitable[DataModelType]]
    """ The function that is used to update the data. """

    trigger: BaseTrigger
    """ The cron trigger that is used to update the data source. """

    _encoder: Encoder | None = None
    """ The encoder that is used to encode and decode the data. """

    _cached_data: DataModelType | None = None
    """ Latest cached data, None if no data has been cached yet. """

    _cache_backend: CacheBackend | None = None
    """ The cache backend that is used to store the data. """

    def __post_init__(self) -> None:
        """Based on the type hints for the function, determine the encoder."""
        return_type = self.function.__annotations__["return"]
        if issubclass(return_type, pl.LazyFrame) or return_type == pl.LazyFrame:
            self._encoder = PolarsEncoder()
        elif issubclass(return_type, BaseModel):
            self._encoder = PydanticEncoder(base_model=return_type)
        else:
            msg = f"No encoder found for type {return_type}"
            raise RuntimeError(msg)

    def set_cache_backend(self, cache_backend: CacheBackend) -> None:
        """Set the cache backend that is used to store the data."""
        self._cache_backend = cache_backend

    async def update(self) -> None:
        """Call the fetch function and update the cache accordingly."""
        if not self._cache_backend:
            msg = "Cache backend not set"
            raise RuntimeError(msg)

        self._cached_data = await self.function(self._cached_data)

        await self._cache_backend.set(
            key=self.name, value=self._encoder.encode(self._cached_data)
        )

    async def load(self) -> None:
        """Load the data from the cache backend."""
        if not self._cache_backend:
            msg = "Cache backend not set"
            raise RuntimeError(msg)

        cache_data = await self._cache_backend.get(key=self.name)
        if cache_data is None:
            return
        self._cached_data = self._encoder.decode(cache_data)

    def get_latest_data(self) -> DataModelType | None:
        """Get the latest data from the data source."""
        data = self._cached_data
        if data is None:
            return None
        return data


# Scheduler


@dataclass
class DataSourceManager:
    """A manager that can schedule data sources to be updated at a specific time."""

    cache_backend: CacheBackend = field(default_factory=SQLiteCache)
    """ The cache backend that is used to store the data. """

    _data_sources: list[CachedDataSource] = field(default_factory=list)
    """ The data sources that are scheduled to be updated. """

    _scheduler: AsyncIOScheduler = field(default_factory=AsyncIOScheduler)
    """ The scheduler that is used to schedule the tasks. """

    # Data Source Scheduling

    def _schedule_data_source(self, data_source: CachedDataSource) -> None:
        """Schedule a data source to be updated according to its trigger.

        ### Arguments
        - data_source: The data source to schedule.
        """

        async def update_job() -> None:
            await data_source.update()

        self._scheduler.add_job(
            update_job,
            trigger=data_source.trigger,
            id=f"update_{data_source.name}",
            replace_existing=True,
        )

    def add_data_source(self, data_source: CachedDataSource) -> None:
        """Add a data source to the scheduler."""
        data_source.set_cache_backend(self.cache_backend)
        self._data_sources.append(data_source)
        self._schedule_data_source(data_source)

    # Lifecycle

    async def start(self) -> None:
        """Start the scheduler."""
        for data_source in self._data_sources:
            await data_source.load()
        self._scheduler.start()

    async def stop(self) -> None:
        """Stop the scheduler."""
        self._scheduler.shutdown()
        await self.cache_backend.cleanup()



"""Models for data models."""

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any

from pandera.dtypes import DataType, is_numeric, is_string
from pandera.polars import DataFrameModel
from pydantic import BaseModel, Field, create_model


class ValueTypeEnum(str, Enum):
    """The type of a value in a column."""

    STRING = "string"
    NUMBER = "number"

    @classmethod
    def from_pandera_dtype(cls, dtype: DataType) -> "ValueTypeEnum":
        """Convert a pandera dtype to a ValueTypeEnum."""
        if is_numeric(dtype):
            return cls.NUMBER
        if is_string(dtype):
            return cls.STRING
        msg = f"Unsupported dtype: {dtype}"
        raise ValueError(msg)

    def into_python_type(self) -> type[Any]:
        """Convert a ValueTypeEnum to a Python type."""
        match self:
            case self.STRING:
                return str
            case self.NUMBER:
                return float
            case _:
                msg = f"Unsupported value type: {self}"
                raise ValueError(msg)


@dataclass
class TabularDatasetColumnSchema:
    """Individual column definition for a tabular dataset."""

    name: str
    """The name of the column."""

    value_type: ValueTypeEnum
    """The type of the values in the column."""


def _create_pydantic_model_from_dataframe(
    model: type[DataFrameModel],
) -> type[BaseModel]:
    """Create a Pydantic model from a DataFrameModel schema.

    ### Arguments:
    - model: The DataFrameModel class to convert.

    ### Returns:
    A dynamically created Pydantic model with the same fields.
    """
    schema = model.to_schema()
    fields: dict[str, tuple[type[Any], Any]] = {}
    for col_name, col in schema.columns.items():
        dtype: DataType = col.dtype
        col.description
        # Convert Pandera types to Python types
        python_type = ValueTypeEnum.from_pandera_dtype(col.dtype).into_python_type()
        # Add field with description if available
        field_kwargs = {}
        if col.description:
            field_kwargs["description"] = col.description
        if col.nullable:
            field_kwargs["default"] = None
            python_type = type(None) | python_type

        # Convert Pandera types to Python types that Pydantic understands
        if hasattr(python_type, "__name__") and python_type.__name__ == "String":
            python_type = str
        elif hasattr(python_type, "__name__") and python_type.__name__ == "Int64":
            python_type = int
        elif hasattr(python_type, "__name__") and python_type.__name__ == "Float64":
            python_type = float
        elif hasattr(python_type, "__name__") and python_type.__name__ == "Boolean":
            python_type = bool

        fields[col_name] = (python_type, Field(**field_kwargs))

    return create_model(f"{model.__name__}Response", **fields)


class TabularViewDataModel(BaseModel):
    """A base model for a tabular dataset."""

    column_schema: BaseModel
    """The schema of the columns of the dataset."""


ViewDataModel = BaseModel | TabularViewDataModel
"""A valid view data model can either be a Pydantic `BaseModel` or a
`TabularViewDataModel`, which itself contains a list of `BaseModel`s.
"""


class Dataset(BaseModel):
    """A base model for a dataset."""

    dataset_id: str
    """The ID of the dataset."""

    last_updated: datetime
    """The last updated date of the dataset."""

    data: ViewDataModel
    """The data of the dataset."""



"""TacoBI is a framework for building and running BI applications."""

from tacobi.bi_app import TacoBIApp
from tacobi.view.view_manager import ViewManager

__all__ = ["TacoBIApp", "ViewManager", "DataSource"]



"""Materialized views."""

from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Generic

from tacobi.data_model.models import DataModelType
from tacobi.data_model.view_data import ViewDataModel
from tacobi.view.base import BaseView


@dataclass
class MaterializedView(BaseView, Generic[DataModelType]):
    """A materialized view."""

    function: Callable[[], Awaitable[DataModelType]]
    """The function to call to update the view."""

    _latest_data: DataModelType | None = None
    """The latest data from the view."""

    def __str__(self) -> str:
        """Get the string representation of the view."""
        return f"MaterializedView(name={self.name}, id={self.id})"

    def get_latest_data(self) -> DataModelType | None:
        """Get the latest data from the view."""
        return self._latest_data

    def get_latest_data_as_base_model(self) -> ViewDataModel:
        """Get the latest data from the view as a BaseModel."""
        return self.transform_to_base_model(self._latest_data)

    async def recompute_latest_data(self) -> None:
        """Recompute the latest data from the view."""
        self._latest_data = await self.function()

    def __hash__(self) -> int:
        """Hash the view."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Check if the view is equal to another object."""
        if not isinstance(other, MaterializedView):
            return False
        return self.id == other.id



"""Dataset schema classes."""

from collections.abc import Callable, Iterable
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Literal, TypeVar

from pandera import Column, DataFrameModel
from pandera.dtypes import DataType, is_numeric, is_string
from pandera.typing.common import DataFrameBase
from pydantic import BaseModel, Field, create_model

DataFrameBaseT = TypeVar("DataFrameBaseT", bound=DataFrameBase)
DatasetFunctionType = Callable[[], DataFrameBaseT]


class DatasetTypeEnum(str, Enum):
    """The type of a dataset."""

    TABULAR = "tabular"
    OBJECT = "object"


class ViewTypeEnum(str, Enum):
    """The type of a view."""

    MATERIALIZED = "materialized"
    NORMAL = "normal"


class ValueTypeEnum(str, Enum):
    """The type of a value in a column. Used for fabricating Pydantic models."""

    STRING = "string"
    NUMBER = "number"

    @classmethod
    def from_pandera_dtype(cls, dtype: DataType) -> "ValueTypeEnum":
        """Convert a pandera dtype to a ValueTypeEnum."""
        if is_numeric(dtype):
            return cls.NUMBER
        if is_string(dtype):
            return cls.STRING
        msg = f"Unsupported dtype: {dtype}"
        raise ValueError(msg)

    def into_python_type(self) -> type[str] | type[float]:
        """Convert a ValueTypeEnum to a Python type."""
        match self:
            case self.STRING:
                return str
            case self.NUMBER:
                return float
            case _:
                msg = f"Unsupported value type: {self}"
                raise ValueError(msg)


def create_column_base_model_from_dataframe_model(
    dataframe_model: type[DataFrameModel],
) -> type[BaseModel]:
    """Create a BaseModel from a DataFrameModel.

    ### Arguments:
    - dataframe_model: The DataFrameModel to create a BaseModel from.

    ### Returns:
    The BaseModel class.
    """
    schema_columns: Iterable[Column] = dataframe_model.to_schema().columns.values()
    fields: dict[str, tuple[type[Any], Any]] = {}
    for col in schema_columns:
        python_type = ValueTypeEnum.from_pandera_dtype(col.dtype).into_python_type()
        fields[col.name] = (python_type, Field(description=col.description))

    return create_model(f"{dataframe_model.__name__}Response", **fields)


class ViewEndpointResponseModel(BaseModel):
    """The return object of a view endpoint."""

    last_updated: datetime
    """The last updated date of the view."""

    view_type: ViewTypeEnum
    """The type of the view. Can be either `MATERIALIZED` or `NORMAL`."""

    data_type: DatasetTypeEnum
    """The type of the data. Can be either `TABULAR` or `OBJECT`."""

    data: BaseModel
    """The data of the view. Always a Pydantic `BaseModel`."""


@dataclass
class ViewEndpointSchema:
    """The schema of a view endpoint. Used for code gen on the frontend."""

    route: str
    """The route of the view endpoint."""

    data_schema: type[BaseModel]
    """The schema of the data. Always a Pydantic `BaseModel`."""

    input_schema: type[BaseModel] | None = None
    """The schema of the input parameters. Always a Pydantic `BaseModel`.
    If the the view is materialized, this is always `None`.
    """

    @property
    def json_schema(self) -> dict[str, Any]:
        """The JSON schema of the data."""
        return {
            "route": self.route,
            "data_schema": self.data_schema.model_json_schema(),
            "input_schema": self.input_schema.model_json_schema()
            if self.input_schema
            else None,
        }


@dataclass
class Schema:
    """The schema of a view."""

    views: list[ViewEndpointSchema]
    """The views in the schema."""

    @property
    def json_schema(self) -> dict[str, Any]:
        """The JSON schema of the schema."""
        return {
            "views": [view.json_schema for view in self.views],
        }



"""Base class for cache backends."""

from abc import ABC, abstractmethod
from dataclasses import dataclass

from tacobi.data_source.encode import EncodedDataType


@dataclass
class CacheBackend(ABC):
    """A cache backend that can be used to store and retrieve data."""

    @abstractmethod
    async def get(self, key: str) -> EncodedDataType | None:
        """Get the data for the given key.

        ### Arguments
        - key: The key to get the data for.

        ### Returns
        - The data for the given key, or None if the data is not found.
        """
        ...

    @abstractmethod
    async def set(self, key: str, value: EncodedDataType) -> None:
        """Set the data for the given key.

        ### Arguments
        - key: The key to set the data for.
        - data: The data to set.
        """
        ...

    @abstractmethod
    async def clear(self) -> None:
        """Clear the cache."""
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Cleanup the cache."""
        ...



"""Materialized and normal views that can be used to query cached data."""

from tacobi.view.base import BaseView
from tacobi.view.materialized_view import MaterializedView
from tacobi.view.view import View
from tacobi.view.view_manager import ViewManager

__all__ = ["BaseView", "MaterializedView", "View", "ViewManager"]



"""Models for data sources."""

from abc import ABC, abstractmethod

from tacobi.data_model.models import DataModelType

EncodedDataType = bytes


class Encoder(ABC):
    """An encoder that can encode and decode data."""

    @abstractmethod
    def encode(self, data: DataModelType) -> EncodedDataType:
        """Encode the data."""
        ...

    @abstractmethod
    def decode(self, data: EncodedDataType) -> DataModelType:
        """Decode the data."""
        ...



"""Views."""

from dataclasses import dataclass
from typing import Generic

from tacobi.data_model.models import DataModelType
from tacobi.view.base import BaseView


@dataclass
class View(BaseView, Generic[DataModelType]):
    """A standard view."""



"""SQLite cache backend."""

import sqlite3 as sql
from dataclasses import dataclass, field
from pathlib import Path

from tacobi.data_source.cache import CacheBackend
from tacobi.data_source.encode import EncodedDataType


def default_db_path() -> Path:
    """Generate a default path to the SQLite database."""
    return Path.cwd() / "cache.db"


@dataclass
class SQLiteCache(CacheBackend):
    """A cache backend that uses a SQLite database to store and retrieve data."""

    db_path: Path = field(default_factory=default_db_path)
    """The path to the SQLite database."""

    _conn: sql.Connection | None = None
    """The SQLite connection."""

    # Init and cleanup

    def __post_init__(self) -> None:
        """Initialize the SQLite connection.

        Create the cache table if it doesn't exist.
        """
        self._conn = sql.connect(self.db_path)
        self._conn.row_factory = sql.Row

        cursor = self._conn.cursor()
        cursor.execute(
            """CREATE TABLE IF NOT EXISTS cache (
            key TEXT PRIMARY KEY,
            data BLOB
        )"""
        )

    def __del__(self) -> None:
        """Close the SQLite connection."""
        self.close()

    # Cache Operations

    async def get(self, key: str) -> EncodedDataType | None:
        """Get the data for the given key.

        ### Arguments
        - key: The key to get the data for.

        ### Returns
        - The data for the given key, or None if the data is not found.
        """
        if not self._conn:
            msg = "SQLite connection not initialized"
            raise RuntimeError(msg)

        cursor = self._conn.cursor()
        cursor.execute("SELECT data FROM cache WHERE key = ?", (key,))
        result = cursor.fetchone()
        return result[0] if result else None

    async def set(self, key: str, value: EncodedDataType) -> None:
        """Set the data for the given key.

        ### Arguments
        - key: The key to set the data for.
        - data: The data to set.
        """
        if not self._conn:
            msg = "SQLite connection not initialized"
            raise RuntimeError(msg)

        cursor = self._conn.cursor()
        cursor.execute(
            """
            INSERT INTO cache (key, data)
            VALUES (?, ?)
            ON CONFLICT(key) DO UPDATE SET data = excluded.data
        """,
            (key, value),
        )
        self._conn.commit()

    async def clear(self) -> None:
        """Clear the cache."""
        if not self._conn:
            msg = "SQLite connection not initialized"
            raise RuntimeError(msg)

        cursor = self._conn.cursor()
        cursor.execute("DELETE FROM cache")
        self._conn.commit()

    def close(self) -> None:
        """Close the SQLite connection."""
        if self._conn:
            self._conn.close()

    async def cleanup(self) -> None:
        """Cleanup the cache."""
        self.close()
